# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

"""Provides LocalAccount class for holding local account private key.

LocalAccount provides operations we need for creating auth key, account address and signing
raw transaction.
"""

from . import (
    diem_types,
    utils,
)

from .auth_key import AuthKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey, Ed25519PublicKey


class LocalAccount:
    """LocalAccount is like a wallet account

    WARN: This is handy class for creating tests for your application, but may not ideal for your
    production code, because it uses a specific implementaion of ed25519 and requires loading your
    private key into memory and hand over to code from external.
    You should always choose more secure way to handle your private key
    (e.g. https://en.wikipedia.org/wiki/Hardware_security_module) in production and do not give
    your private key to any code from external if possible.
    """

    @staticmethod
    def generate() -> "LocalAccount":
        """Generate a random private key and initialize local account"""

        private_key = Ed25519PrivateKey.generate()
        return LocalAccount(private_key)

    @staticmethod
    def from_private_key_hex(key: str) -> "LocalAccount":
        return LocalAccount(Ed25519PrivateKey.from_private_bytes(bytes.fromhex(key)))

    private_key: Ed25519PrivateKey
    compliance_key: Ed25519PrivateKey

    def __init__(self, private_key: Ed25519PrivateKey) -> None:
        self.private_key = private_key
        self.compliance_key = Ed25519PrivateKey.generate()

    @property
    def auth_key(self) -> AuthKey:
        return AuthKey.from_public_key(self.public_key)

    @property
    def account_address(self) -> diem_types.AccountAddress:
        return self.auth_key.account_address()

    @property
    def public_key_bytes(self) -> bytes:
        return utils.public_key_bytes(self.public_key)

    @property
    def public_key(self) -> Ed25519PublicKey:
        return self.private_key.public_key()

    @property
    def compliance_public_key_bytes(self) -> bytes:
        return utils.public_key_bytes(self.compliance_key.public_key())

    def sign(self, txn: diem_types.RawTransaction) -> diem_types.SignedTransaction:
        """Create signed transaction for given raw transaction"""

        signature = self.private_key.sign(utils.raw_transaction_signing_msg(txn))
        return utils.create_signed_transaction(txn, self.public_key_bytes, signature)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

import dataclasses, typing, abc


S = typing.TypeVar("S")
T = typing.TypeVar("T")


@dataclasses.dataclass(frozen=True)
class MatchResult:
    success: bool = dataclasses.field(default=False)
    matched_fields: typing.List[str] = dataclasses.field(default_factory=lambda: [])
    mismatched_fields: typing.List[str] = dataclasses.field(default_factory=lambda: [])

    @staticmethod
    def create(success: bool, fields: typing.List[str]) -> "MatchResult":
        return MatchResult(
            success=success,
            matched_fields=fields if success else [],
            mismatched_fields=fields if not success else [],
        )

    @staticmethod
    def merge(ret1: "MatchResult", ret2: "MatchResult") -> "MatchResult":
        return dataclasses.replace(
            ret1,
            success=ret1.success and ret2.success,
            matched_fields=ret1.matched_fields + ret2.matched_fields,
            mismatched_fields=ret1.mismatched_fields + ret2.mismatched_fields,
        )


class Condition(abc.ABC, typing.Generic[T]):
    @abc.abstractmethod
    def match(self, event_data: T) -> MatchResult:
        ...


@dataclasses.dataclass(frozen=True)
class Field(Condition[T]):
    path: str
    not_set: bool = dataclasses.field(default=False)

    def match(self, event_data: T) -> MatchResult:
        val = event_data
        for f in self.path.split("."):
            if val is None or not hasattr(val, f):
                return MatchResult.create(False, [self.path])
            val = getattr(val, f)

        if self.not_set:
            return MatchResult.create(val is None, [self.path])
        return MatchResult.create(val is not None, [self.path])


@dataclasses.dataclass(frozen=True)
class Value(typing.Generic[T, S], Condition[T]):
    path: str
    value: S

    def match(self, event_data: T) -> MatchResult:
        val = event_data
        for f in self.path.split("."):
            if val is None or not hasattr(val, f):
                return MatchResult.create(False, [self.path])
            val = getattr(val, f)
        return MatchResult.create(val == self.value, [self.path])


class ConditionValidationError(Exception):
    def __init__(self, validation: Condition[T], match_result: MatchResult) -> None:
        super().__init__(f"mismatch result: {match_result}")
        self.validation = validation
        self.match_result = match_result


@dataclasses.dataclass(frozen=True)
class Require(Condition[T]):
    conds: typing.List[Condition[T]]
    validation: typing.Optional[Condition[T]] = dataclasses.field(default=None)

    def match(self, event_data: T) -> MatchResult:
        ret = MatchResult(success=True)
        for cond in self.conds:
            ret = MatchResult.merge(ret, cond.match(event_data))

        if ret.success and self.validation:
            ret = MatchResult.merge(ret, self.validation.match(event_data))
            if not ret.success:
                raise ConditionValidationError(self.validation, ret)

        return ret

    def __hash__(self) -> int:
        return hash(tuple(self.conds))


@dataclasses.dataclass(frozen=True)
class State(typing.Generic[T]):
    id: str
    require: typing.Optional[Require[T]] = dataclasses.field(default=None)

    def match(self, event_data: T) -> MatchResult:
        if self.require:
            return self.require.match(event_data)
        return MatchResult(success=True)

    def __str__(self) -> str:
        return self.id


@dataclasses.dataclass(frozen=True)
class Transition(typing.Generic[T]):
    action: str
    state: State[T]
    to: State[T]


class NoStateMatchedError(ValueError):
    pass


class TooManyStatesMatchedError(ValueError):
    pass


@dataclasses.dataclass
class Machine(typing.Generic[T]):
    initials: typing.List[State[T]]
    states: typing.List[State[T]]
    transitions: typing.List[Transition[T]]

    def is_initial(self, state: State[T]) -> bool:
        return state in self.initials

    def is_valid_transition(self, state: State[T], to: State[T], event_data: T) -> bool:
        for t in self.transitions:
            if t.state == state and t.to == to:
                return True
        return False

    def match_state(self, event_data: T) -> State[T]:
        ret = self.match_states(event_data)
        if not ret:
            raise NoStateMatchedError(f"could not find state matches given event data({event_data})")
        if len(ret) > 1:
            raise TooManyStatesMatchedError(f"found multiple states({ret}) match given event data({event_data})")
        return ret[0]

    def match_states(self, event_data: T) -> typing.List[State[T]]:
        return [state for state, match in self.match_states_and_results(event_data) if match.success]

    def match_states_and_results(self, event_data: T) -> typing.List[typing.Tuple[State[T], MatchResult]]:
        return [(state, state.match(event_data)) for state in self.states]


def new_transition(state: State[T], to: State[T]) -> Transition[T]:
    return Transition(action=f"{state} -> {to}", state=state, to=to)


def require(*args: Condition[T], validation: typing.Optional[Condition[T]] = None) -> Require[T]:
    return Require(conds=list(args), validation=validation)


def build_machine(transitions: typing.List[Transition[T]]) -> Machine[T]:
    states = {}
    tos = {}
    for t in transitions:
        states[t.state.id] = t.state
        states[t.to.id] = t.to
        tos[t.to.id] = t.to

    initial_ids = set(states.keys()) - set(tos.keys())
    return Machine(
        initials=[states[id] for id in initial_ids],
        states=list(states.values()),
        transitions=transitions,
    )



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

import requests, typing, dataclasses, uuid

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.exceptions import InvalidSignature
from json.decoder import JSONDecodeError

from .payment_command import PaymentCommand

from .types import (
    CommandType,
    CommandRequestObject,
    CommandResponseObject,
    CommandResponseStatus,
    PaymentObject,
    PaymentActorObject,
    PaymentActionObject,
    PaymentCommandObject,
    ErrorCode,
    FieldError,
)
from .error import command_error, protocol_error, Error

from . import jws, http_header
from .. import jsonrpc, diem_types, identifier, utils


DEFAULT_CONNECT_TIMEOUT_SECS: float = 2.0
DEFAULT_TIMEOUT_SECS: float = 5.0


class CommandResponseError(Exception):
    def __init__(self, resp: CommandResponseObject) -> None:
        super().__init__(resp)
        self.resp = resp


@dataclasses.dataclass
class Client:
    my_compliance_key_account_address: diem_types.AccountAddress
    jsonrpc_client: jsonrpc.Client
    hrp: str
    session: requests.Session = dataclasses.field(default_factory=lambda: requests.Session())
    timeout: typing.Tuple[float, float] = dataclasses.field(
        default_factory=lambda: (
            DEFAULT_CONNECT_TIMEOUT_SECS,
            DEFAULT_TIMEOUT_SECS,
        )
    )
    my_compliance_key_account_id: str = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        self.my_compliance_key_account_id = self.account_id(self.my_compliance_key_account_address)

    def send_command(self, command: PaymentCommand, sign: typing.Callable[[bytes], bytes]) -> CommandResponseObject:
        return self.send_request(
            request_sender_address=command.my_actor_address,
            opponent_account_id=command.opponent_actor_obj().address,
            request_bytes=jws.serialize(command.new_request(), sign),
        )

    def send_request(
        self, request_sender_address: str, opponent_account_id: str, request_bytes: bytes
    ) -> CommandResponseObject:
        base_url, public_key = self.get_base_url_and_compliance_key(opponent_account_id)
        response = self.session.post(
            f"{base_url.rstrip('/')}/v2/command",
            data=request_bytes,
            headers={
                http_header.X_REQUEST_ID: str(uuid.uuid4()),
                http_header.X_REQUEST_SENDER_ADDRESS: request_sender_address,
            },
            timeout=self.timeout,
        )
        if response.status_code not in [200, 400]:
            response.raise_for_status()

        cmd_resp = _deserialize_jws(response.content, CommandResponseObject, public_key, protocol_error)
        if cmd_resp.status == CommandResponseStatus.failure:
            raise CommandResponseError(cmd_resp)
        return cmd_resp

    def process_inbound_request(self, request_sender_address: str, request_bytes: bytes) -> PaymentCommand:
        if not request_sender_address:
            raise protocol_error(ErrorCode.missing_http_header, f"missing {http_header.X_REQUEST_SENDER_ADDRESS}")
        try:
            _, public_key = self.get_base_url_and_compliance_key(request_sender_address)
        except ValueError as e:
            raise protocol_error(ErrorCode.invalid_x_request_sender_address, str(e)) from e

        request = _deserialize_jws(request_bytes, CommandRequestObject, public_key, command_error)
        if request.command_type == CommandType.PaymentCommand:
            payment = typing.cast(PaymentCommandObject, request.command).payment
            self.validate_addresses(payment, request_sender_address)
            cmd = self.create_inbound_payment_command(request.cid, payment)
            if cmd.is_initial():
                self.validate_dual_attestation_limit(cmd.payment.action)
            elif cmd.is_rsend():
                self.validate_recipient_signature(cmd, public_key)
            return cmd

        raise command_error(
            ErrorCode.unknown_command_type,
            f"unknown command_type: {request.command_type}",
        )

    def validate_recipient_signature(self, cmd: PaymentCommand, public_key: Ed25519PublicKey) -> None:
        msg = cmd.travel_rule_metadata_signature_message(self.hrp)
        try:
            sig = cmd.payment.recipient_signature
            if sig is None:
                raise ValueError("recipient_signature is not provided")
            public_key.verify(bytes.fromhex(sig), msg)
        except (ValueError, InvalidSignature) as e:
            raise command_error(
                ErrorCode.invalid_recipient_signature, str(e), "command.payment.recipient_signature"
            ) from e

    def validate_dual_attestation_limit(self, action: PaymentActionObject) -> None:
        currencies = self.jsonrpc_client.get_currencies()
        limit = self.jsonrpc_client.get_metadata().dual_attestation_limit
        for info in currencies:
            if info.code == action.currency:
                approx_xdx_microdiem_value = info.to_xdx_exchange_rate * action.amount
                if approx_xdx_microdiem_value < limit:
                    raise command_error(
                        ErrorCode.no_kyc_needed,
                        f"payment amount is under travel rule threshold {limit}",
                        "command.payment.action.amount",
                    )
                return
        raise command_error(
            ErrorCode.invalid_field_value,
            f"currency code is invalid: {action.currency}",
            "command.payment.action.currency",
        )

    def validate_addresses(self, payment: PaymentObject, request_sender_address: str) -> None:
        self.validate_actor_address("sender", payment.sender)
        self.validate_actor_address("receiver", payment.receiver)
        self.validate_request_sender_address(request_sender_address, [payment.sender.address, payment.receiver.address])

    def validate_actor_address(self, actor_name: str, actor: PaymentActorObject) -> None:
        try:
            identifier.decode_account(actor.address, self.hrp)
        except ValueError as e:
            raise command_error(
                ErrorCode.invalid_field_value,
                f"could not decode account identifier: {e}",
                f"command.payment.{actor_name}.address",
            ) from e

    def validate_request_sender_address(self, request_sender_address: str, addresses: typing.List[str]) -> None:
        if request_sender_address not in addresses:
            raise command_error(
                ErrorCode.invalid_x_request_sender_address,
                f"address {request_sender_address} is not one of {addresses}",
            )

    def create_inbound_payment_command(self, cid: str, obj: PaymentObject) -> PaymentCommand:
        if self.is_my_account_id(obj.sender.address):
            return PaymentCommand(cid=cid, my_actor_address=obj.sender.address, payment=obj, inbound=True)
        if self.is_my_account_id(obj.receiver.address):
            return PaymentCommand(cid=cid, my_actor_address=obj.receiver.address, payment=obj, inbound=True)

        raise command_error(ErrorCode.unknown_actor_address, "unknown actor addresses: {obj}")

    def is_my_account_id(self, account_id: str) -> bool:
        account_address, _ = identifier.decode_account(account_id, self.hrp)
        if self.my_compliance_key_account_id == self.account_id(account_address):
            return True
        account = self.jsonrpc_client.get_account(account_address)
        if account and account.role.parent_vasp_address:
            return self.my_compliance_key_account_id == self.account_id(account.role.parent_vasp_address)
        return False

    def account_id(self, address: typing.Union[diem_types.AccountAddress, bytes, str]) -> str:
        return identifier.encode_account(utils.account_address(address), None, self.hrp)

    def get_base_url_and_compliance_key(self, account_id: str) -> typing.Tuple[str, Ed25519PublicKey]:
        account_address, _ = identifier.decode_account(account_id, self.hrp)
        return self.jsonrpc_client.get_base_url_and_compliance_key(account_address)


def _deserialize_jws(
    content_bytes: bytes,
    klass: typing.Type[jws.T],
    public_key: Ed25519PublicKey,
    error_fn: typing.Callable[[str, str, typing.Optional[str]], Error],
) -> jws.T:
    try:
        return jws.deserialize(content_bytes, klass, public_key.verify)
    except JSONDecodeError as e:
        raise error_fn(ErrorCode.invalid_json, f"decode json string failed: {e}", None) from e
    except FieldError as e:
        raise error_fn(e.code, f"invalid {klass.__name__} json: {e}", e.field) from e
    except InvalidSignature as e:
        raise error_fn(ErrorCode.invalid_jws_signature, f"invalid jws signature: {e}", None) from e
    except ValueError as e:
        raise error_fn(ErrorCode.invalid_jws, f"deserialize JWS bytes failed: {e}", None) from e



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

""" This package provides data structures and utilities for implementing Diem Offchain API Service.

See [Diem Offchain API](https://dip.diem.com/dip-1/) for more details.

"""

from .types import (
    AbortCode,
    CommandType,
    CommandResponseStatus,
    OffChainErrorType,
    ErrorCode,
    PaymentCommandObject,
    CommandRequestObject,
    CommandResponseObject,
    OffChainErrorObject,
    PaymentObject,
    PaymentActorObject,
    PaymentActionObject,
    Status,
    StatusObject,
    KycDataObjectType,
    NationalIdObject,
    AddressObject,
    KycDataObject,
    FieldError,
    InvalidOverwriteError,
    new_payment_object,
    new_payment_request,
    replace_payment_actor,
    reply_request,
    individual_kyc_data,
    entity_kyc_data,
    to_json,
    from_json,
    validate_write_once_fields,
)
from .http_header import X_REQUEST_ID, X_REQUEST_SENDER_ADDRESS
from .action import Action
from .error import command_error, protocol_error, Error
from .command import Command
from .payment_command import PaymentCommand
from .client import Client, CommandResponseError

from . import jws, http_server, state, payment_state

import typing



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

"""Provides utilities for working with Diem Testnet.

```python

from diem import testnet, LocalAccount

# create client connects to testnet
client = testnet.create_client()

# create faucet for minting coins for your testing account
faucet = testnet.Faucet(client)

# create a local account and mint some coins for it
account: LocalAccount = faucet.gen_account()

```

"""

import requests
import typing, time

from . import diem_types, jsonrpc, utils, chain_ids, bcs, stdlib, LocalAccount


JSON_RPC_URL: str = "http://testnet.diem.com/v1"
FAUCET_URL: str = "http://testnet.diem.com/mint"
CHAIN_ID: diem_types.ChainId = chain_ids.TESTNET

DESIGNATED_DEALER_ADDRESS: diem_types.AccountAddress = utils.account_address("000000000000000000000000000000dd")
TEST_CURRENCY_CODE: str = "XUS"


def create_client() -> jsonrpc.Client:
    """create a jsonrpc.Client connects to Testnet public full node cluster"""

    return jsonrpc.Client(JSON_RPC_URL)


def gen_vasp_account(client: jsonrpc.Client, base_url: str) -> LocalAccount:
    account = gen_account(client)
    exec_txn(
        client,
        account,
        stdlib.encode_rotate_dual_attestation_info_script(
            new_url=base_url.encode("utf-8"), new_key=account.compliance_public_key_bytes
        ),
    )
    return account


def gen_account(client: jsonrpc.Client, dd_account: bool = False) -> LocalAccount:
    """generates a Testnet onchain account"""

    return Faucet(client).gen_account(dd_account=dd_account)


def gen_child_vasp(
    client: jsonrpc.Client,
    parent_vasp: LocalAccount,
    initial_balance: int = 10_000_000_000,
    currency: str = TEST_CURRENCY_CODE,
) -> LocalAccount:
    child_vasp = LocalAccount.generate()
    exec_txn(
        client,
        parent_vasp,
        stdlib.encode_create_child_vasp_account_script(
            coin_type=utils.currency_code(currency),
            child_address=child_vasp.account_address,
            auth_key_prefix=child_vasp.auth_key.prefix(),
            add_all_currencies=False,
            child_initial_balance=initial_balance,
        ),
    )
    return child_vasp


def exec_txn(client: jsonrpc.Client, sender: LocalAccount, script: diem_types.Script) -> jsonrpc.Transaction:
    """create, submit and wait for the transaction"""

    sender_account_sequence = client.get_account_sequence(sender.account_address)
    expire = 30
    txn = sender.sign(
        diem_types.RawTransaction(  # pyre-ignore
            sender=sender.account_address,
            sequence_number=diem_types.st.uint64(sender_account_sequence),
            payload=diem_types.TransactionPayload__Script(value=script),
            max_gas_amount=1_000_000,
            gas_unit_price=0,
            gas_currency_code=TEST_CURRENCY_CODE,
            expiration_timestamp_secs=int(time.time()) + expire,
            chain_id=CHAIN_ID,
        )
    )
    client.submit(txn)
    return client.wait_for_transaction(txn, timeout_secs=expire)


class Faucet:
    """Faucet service is a proxy server to mint coins for your test account on Testnet

    See https://github.com/diem/diem/blob/master/json-rpc/docs/service_testnet_faucet.md for more details
    """

    def __init__(
        self,
        client: jsonrpc.Client,
        url: typing.Union[str, None] = None,
        retry: typing.Union[jsonrpc.Retry, None] = None,
    ) -> None:
        self._client: jsonrpc.Client = client
        self._url: str = url or FAUCET_URL
        self._retry: jsonrpc.Retry = retry or jsonrpc.Retry(5, 0.2, Exception)
        self._session: requests.Session = requests.Session()

    def gen_account(self, currency_code: str = TEST_CURRENCY_CODE, dd_account: bool = False) -> LocalAccount:
        account = LocalAccount.generate()
        self.mint(account.auth_key.hex(), 100_000_000_000, currency_code, dd_account)
        return account

    def mint(self, authkey: str, amount: int, currency_code: str, dd_account: bool = False) -> None:
        self._retry.execute(lambda: self._mint_without_retry(authkey, amount, currency_code, dd_account))

    def _mint_without_retry(self, authkey: str, amount: int, currency_code: str, dd_account: bool = False) -> None:
        response = self._session.post(
            FAUCET_URL,
            params={
                "amount": amount,
                "auth_key": authkey,
                "currency_code": currency_code,
                "return_txns": "true",
                "is_designated_dealer": "true" if dd_account else "false",
            },
        )
        response.raise_for_status()

        de = bcs.BcsDeserializer(bytes.fromhex(response.text))
        length = de.deserialize_len()

        for i in range(length):
            txn = de.deserialize_any(diem_types.SignedTransaction)
            self._client.wait_for_transaction(txn)



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

""" This module defines util functions for testing offchain inbound request processor with http server

Not recommended for production
"""

from http import server
from .http_header import X_REQUEST_ID, X_REQUEST_SENDER_ADDRESS

import logging, socket, threading, typing

logger: logging.Logger = logging.getLogger(__name__)


def start_local(
    port: int,
    process_inbound_request: typing.Callable[[str, str, bytes], typing.Tuple[int, bytes]],
) -> server.HTTPServer:
    """starts a HTTPServer on localhost with given port and callback function to process offchain inbound requests

    Warning: this is not recommended for production.

    Given process_inbound_request callable will receive:
    1. X_REQUEST_ID header value for logging purpose.
    2. DIP-5 account id for verifying JWS signature.
    3. JWS content bytes from request body.
    """

    class Handler(server.BaseHTTPRequestHandler):
        def do_POST(self):
            x_request_id = self.headers[X_REQUEST_ID]
            jws_key_address = self.headers[X_REQUEST_SENDER_ADDRESS]
            try:
                length = int(self.headers["content-length"])
                content = self.rfile.read(length)

                code, resp_body = process_inbound_request(x_request_id, jws_key_address, content)
                self.send_response(code)
                self.send_header(X_REQUEST_ID, x_request_id)
                self.end_headers()
                self.wfile.write(resp_body)
            except Exception as e:
                logger.exception(e)
                self.send_error(500, str(e))

    httpd = server.HTTPServer(("localhost", port), Handler)

    t = threading.Thread(target=httpd.serve_forever)
    t.daemon = True
    t.start()

    return httpd


def get_available_port() -> int:
    """get_available_port returns an available port"""

    with socket.socket() as s:
        s.bind(("localhost", 0))
        return s.getsockname()[1]



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

"""Defines constants for enum type values

See the following Diem JSON-RPC response type documents for more details:

* https://github.com/diem/diem/blob/master/json-rpc/docs/type_account.md#type-account
* https://github.com/diem/diem/blob/master/json-rpc/docs/type_event.md#event-data
* https://github.com/diem/diem/blob/master/json-rpc/docs/type_transaction.md#type-vmstatus
* https://github.com/diem/diem/blob/master/json-rpc/docs/type_transaction.md#type-transactiondata


"""

# AccountRole#type field values
ACCOUNT_ROLE_UNKNOWN: str = "unknown"
ACCOUNT_ROLE_CHILD_VASP: str = "child_vasp"
ACCOUNT_ROLE_PARENT_VASP: str = "parent_vasp"
ACCOUNT_ROLE_DESIGNATED_DEALER: str = "designated_dealer"


# EventData#type field values
EVENT_DATA_UNKNOWN: str = "unknown"
EVENT_DATA_BURN: str = "burn"
EVENT_DATA_CANCEL_BURN: str = "cancelburn"
EVENT_DATA_MINT: str = "mint"
EVENT_DATA_TO_XDX_EXCHANGE_RATE_UPDATE: str = "to_xdx_exchange_rate_update"
EVENT_DATA_PREBURN: str = "preburn"
EVENT_DATA_RECEIVED_PAYMENT: str = "receivedpayment"
EVENT_DATA_SENT_PAYMENT: str = "sentpayment"
EVENT_DATA_NEW_EPOCH: str = "newepoch"
EVENT_DATA_NEW_BLOCK: str = "newblock"
EVENT_DATA_RECEIVED_MINT: str = "receivedmint"
EVENT_DATA_COMPLIANCE_KEY_ROTATION: str = "compliancekeyrotation"
EVENT_DATA_BASE_URL_ROTATION: str = "baseurlrotation"
EVENT_DATA_CREATE_ACCOUNT: str = "createaccount"
EVENT_DATA_ADMIN_TRANSACTION: str = "admintransaction"


# VMStatus#type field values
VM_STATUS_EXECUTED: str = "executed"
VM_STATUS_OUT_OF_GAS: str = "out_of_gas"
VM_STATUS_MOVE_ABORT: str = "move_abort"
VM_STATUS_EXECUTION_FAILURE: str = "execution_failure"
VM_STATUS_MISC_ERROR: str = "miscellaneous_error"


# TransactionData#type field values
TRANSACTION_DATA_BLOCK_METADATA: str = "blockmetadata"
TRANSACTION_DATA_WRITE_SET: str = "writeset"
TRANSACTION_DATA_USER: str = "user"
TRANSACTION_DATA_UNKNOWN: str = "unknown"


# Script#type field values, only set unknown type here,
# other types, plese see https://github.com/diem/diem/blob/master/language/stdlib/transaction_scripts/doc/transaction_script_documentation.md for all available script names.
SCRIPT_UNKNOWN: str = "unknown"



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

import os
from setuptools import setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="diem",
    version="1.1.3",
    description="The Python Client SDK for Diem",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="Apache-2.0",
    url="https://github.com/diem/client-sdk-python",
    # author="Diem Open Source",
    python_requires=">=3.7", # requires dataclasses
    packages=["diem"],
    # package_data={"": ["src/diem/jsonrpc/*.pyi"]},
    package_dir={"": "src"},
    include_package_data=True,  # see MANIFEST.in
    zip_safe=True,
    install_requires=["requests>=2.20.0", "cryptography>=2.8", "numpy>=1.18", "protobuf>=3.12.4"],
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        "setuptools>=18.0",
    ],
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)



# Copyright (c) Facebook, Inc. and its affiliates
# SPDX-License-Identifier: MIT OR Apache-2.0

import numpy as np
from dataclasses import dataclass
import typing


class SerializationError(ValueError):
    """Error raised during Serialization"""

    pass


class DeserializationError(ValueError):
    """Error raised during Deserialization"""

    pass


@dataclass(init=False)
class uint128:
    high: np.uint64
    low: np.uint64

    def __init__(self, num):
        self.high = np.uint64(num >> 64)
        self.low = np.uint64(num & 0xFFFFFFFFFFFFFFFF)

    def __int__(self):
        return (int(self.high) << 64) | int(self.low)


@dataclass(init=False)
class int128:
    high: np.int64
    low: np.uint64

    def __init__(self, num):
        self.high = np.int64(num >> 64)
        self.low = np.uint64(num & 0xFFFFFFFFFFFFFFFF)

    def __int__(self):
        return (int(self.high) << 64) | int(self.low)


@dataclass(init=False)
class char:
    value: str

    def __init__(self, s):
        if len(s) != 1:
            raise ValueError("`char` expects a single unicode character")
        self.value = s

    def __str__(self):
        return self.value


unit = typing.Type[None]

bool = bool
int8 = np.int8
int16 = np.int16
int32 = np.int32
int64 = np.int64

uint8 = np.uint8
uint16 = np.uint16
uint32 = np.uint32
uint64 = np.uint64

float32 = np.float32
float64 = np.float64



# Copyright (c) The Diem Core Contributors
# SPDX-License-Identifier: Apache-2.0

""" This package provides a client for connecting to Diem JSON-RPC Service API

Create a client connect to Diem Testnet and calls get_metadata API:

```python3

>>> from diem import jsonrpc, testnet
>>> client = jsonrpc.Client(testnet.JSON_RPC_URL)
>>> client.get_metadata()
version: 3300304
timestamp: 1601492912847973
chain_id: 2

```

See [Diem JSON-RPC API SPEC](https://github.com/diem/diem/blob/master/json-rpc/json-rpc-spec.md) for more details

"""

from .client import (
    Client,
    State,
    Retry,
    RequestStrategy,
    RequestWithBackups,
    # Exceptions
    JsonRpcError,
    NetworkError,
    InvalidServerResponse,
    StaleResponseError,
    TransactionHashMismatchError,
    TransactionExecutionFailed,
    TransactionExpired,
    WaitForTransactionTimeout,
    AccountNotFoundError,
)
from .jsonrpc_pb2 import (
    Amount,
    Metadata,
    CurrencyInfo,
    Account,
    AccountRole,
    Transaction,
    TransactionData,
    Script,
    Event,
    EventData,
    VMStatus,
    StateProof,
    AccountStateWithProof,
    AccountStateProof,
)
from .constants import (
    # AccountRole#type field values
    ACCOUNT_ROLE_UNKNOWN,
    ACCOUNT_ROLE_CHILD_VASP,
    ACCOUNT_ROLE_PARENT_VASP,
    ACCOUNT_ROLE_DESIGNATED_DEALER,
    # EventData#type field values
    EVENT_DATA_UNKNOWN,
    EVENT_DATA_BURN,
    EVENT_DATA_CANCEL_BURN,
    EVENT_DATA_MINT,
    EVENT_DATA_TO_XDX_EXCHANGE_RATE_UPDATE,
    EVENT_DATA_PREBURN,
    EVENT_DATA_RECEIVED_PAYMENT,
    EVENT_DATA_SENT_PAYMENT,
    EVENT_DATA_NEW_EPOCH,
    EVENT_DATA_NEW_BLOCK,
    EVENT_DATA_RECEIVED_MINT,
    EVENT_DATA_COMPLIANCE_KEY_ROTATION,
    EVENT_DATA_BASE_URL_ROTATION,
    EVENT_DATA_CREATE_ACCOUNT,
    EVENT_DATA_ADMIN_TRANSACTION,
    # VMStatus#type field values
    VM_STATUS_EXECUTED,
    VM_STATUS_OUT_OF_GAS,
    VM_STATUS_MOVE_ABORT,
    VM_STATUS_EXECUTION_FAILURE,
    VM_STATUS_MISC_ERROR,
    # TransactionData#type field values
    TRANSACTION_DATA_BLOCK_METADATA,
    TRANSACTION_DATA_WRITE_SET,
    TRANSACTION_DATA_USER,
    TRANSACTION_DATA_UNKNOWN,
    # Script#type field values, only set unknown type here,
    # other types, plese see https://github.com/diem/diem/blob/master/language/stdlib/transaction_scripts/doc/transaction_script_documentation.md for all available script names.
    SCRIPT_UNKNOWN,
)
